//
// Copyright (C) 2002 Henning Spruth

// Written by Henning Spruth <henning.spruth@epost.de>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <stdlib.h>
#ifdef DMALLOC
#include <dmalloc.h>
#endif

#include <FL/Fl.H>
#include <FL/fl_draw.H>


#include "FontStack.h"

FontStack::FontStack() {
    stack=(Font *) calloc(increment, sizeof(Font));
    alloc=increment;
    count=0;
}

FontStack::~FontStack() {
    free((void *) stack);
}


void FontStack::push(Font *&current, uchar font, uchar size) {
    if(count>=alloc-1) {
	alloc+=increment;
	stack = (Font *) realloc(stack, sizeof(Font)*alloc);
    }
    stack[count].font=current->font;
    stack[count].size=current->size;
    stack[count].spcWidth=current->spcWidth;
    count++;
    current = &stack[count];
    current->font=font;
    current->size=size;
    fl_font(font,size);
    current->spcWidth=(uchar) fl_width(" ");
}

void FontStack::pop(Font* &current) {
    if(count>0) count--;
    current=&stack[count];
    fl_font(current->font, current->size);
}

void FontStack::clear() {
    count=0;
}
