//
// Copyright (C) 2002 Henning Spruth
// Copyright 1997-2000 by Easy Software Products.

// Written by Henning Spruth <henning.spruth@epost.de>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>

#ifdef DMALLOC
#include <dmalloc.h>
#endif

#include <FL/Fl.H>
//#include <FL/Fl_Group.H>
//#include <FL/Fl_Scrollbar.H>
#include <FL/fl_draw.H>

#include "LinkArray.h"
#include "FontStack.h"
#include "LineDef.h"
#include "Para.h"
#include "ParaStack.h"
#include "Table.h"
#include "TableStack.h"
#include "HtmlData.h"

void HtmlData::loadFile(const char *fileName) {
    FILE *fp;

    clear();

    // free old text buffer
    if(htmlText!=NULL)
	    free((void *) htmlText);

    // Try to open file
    if (fileName!=NULL && (fp = fopen(fileName, "rb")) != NULL)
    {
	// Determine size of file ...
	size_t len;
	fseek(fp, 0, SEEK_END);
	len = ftell(fp);
	rewind(fp);
	
	// ... allocate the buffer accordingly ...
	htmlText = (const char *)calloc(len + 1, 1);

	// .. and read the file
	fread((void *)htmlText, 1, len, fp);
	fclose(fp);
    }
    else
    {
	if(fileName==NULL)
	    htmlText=strdup("ERROR: called with NULL file name");
	else {
	    // Couldn't open input file
	    char error[512];
	    sprintf(error, "%s: %s\n", fileName, strerror(errno));
	    htmlText=strdup(error);
	}
    }
}


#if 0
short HtmlData::newLinkIndex(char *target) {
    if(linkCount>=linkAlloc) {
	linkAlloc+=10;
	if(linkAlloc==10) 
	    links=(Link *) malloc(sizeof(Link) * linkAlloc);
	else 
	    links = (Link *) realloc(links, sizeof(Link)*linkAlloc);
    }
    result = links+linkCount;
    result->target=strdup(target);
    linkCount++;
    return linkCount-1;
}
#endif

Link *HtmlData::getLink(short index) {
    return linkArray->get(index);
}



short HtmlData::getElemMaxSize(Elem *elem) {
    int lineIdx = elem->lineIdx;
    return lineDef->maxSize(lineIdx);

}


//
// Traverse the element list, and fix the x and y position
// of each element.
// 
// The font and widths have already been determined in the
// call to measure()
//
void HtmlData::format(int screenWidth) {

    if(minPageWidth<screenWidth)
	pagew=screenWidth;
    else
	pagew=minPageWidth;
    pageh=0;

    short currentLink=-1;

    Para *p = new Para(lineDef, pagew,0);

    Table *curTable=NULL;
    char *argval;	// scratch for IMG 


    for(int i=0; i<elemCount; i++) {
	short newy;
	Elem *elem=&(elements[i]);
	switch(elem->type) {
	case Tcomment:
	    // ignore comments;
	    break;
	case Tpre:
	    if(p->posx>0) 
		p->newLine();
	    elem->x=p->x+p->posx;
	    elem->y=p->y+p->posy;
	    elem->lineIdx=p->curLineIdx;
	    //fl_font(elem->font, elem->size);
	    elem->h=elem->size;
	    p->maxLineSize(elem->size);
	    p->newLine();
	    break;
	case Timg:
	    // The IMG tag has been converted to be comptabile wit
	    // normal text in the measure function
	case Ttext:
	    elem->h=elem->size;
	    if(p->posx>0) {
		if(p->posx+elem->spcWidth+elem->w > p->w) {
		    p->newLine();
		} else 
		    p->posx+=elem->spcWidth;
	    }
	    p->maxLineSize(elem->size);
	    elem->lineIdx=p->curLineIdx;
	    elem->x=p->x+p->posx;
	    elem->y=p->y+p->posy;
	    p->posx+=elem->w;
	    elem->linkIndex=currentLink;
	    //printf("%d, %d w=%d: %s\n", elem->x, elem->y, elem->w, elem->args);
	    break;
	case Th1:
	case Th2:
	case Th3:
	case Th4:
	case Th5:
	case Th6:
	case Uh1:
	case Uh2:
	case Uh3:
	case Uh4:
	case Uh5:
	case Uh6:
	case Tbr:
	    p->newLine();
	    break;
	case Tp:
	case Utitle:
	    if(p->posx>0) {
		int ms=p->getLineSize();
		p->newLine();
		// make sure empty line has some height
		p->maxLineSize(ms);
	    }
	    p->newLine();
	    break;
	case Ta:
	    argval=getArg(elem, "href");
	    if(argval!=NULL)
		currentLink=linkArray->addLink(argval);
	    break;
	case Ua:
	    currentLink=-1;
	    break;
	case Thr:
	    if(p->posx>0) {
		p->newLine();
	    }
	    // The horizontal line is padded by 3 pixesl on top and bottom
	    // for a total height of 7
	    elem->x=p->x+p->posx;
	    elem->y=p->y+p->posy+4;
	    p->posy+=7;
	    break;
	    //
	    // LISTS
	    //
	case Tul:
	case Tol:
	    if(p->posx>0) 
		p->newLine();
	    paraStack->push(lineDef, p);
	    // leave room for bullet
	    p->x+=5*elem->spcWidth;
	    p->w-=5*elem->spcWidth;
	    break;
	case Tli:
	    if(p->posx>0) 
		p->newLine();
	    elem->x=p->x+p->posx;
	    elem->y=p->y+p->posy;
	    elem->lineIdx=p->curLineIdx;
	    break;
	case Uul:
	case Uol:
	    if(p->posx>0)
		p->newLine();
	    newy=p->posy;
	    paraStack->pop(p);
	    p->posx=0;
	    p->posy+=newy;
	    break;
	    //
	    // TABLES
	    //
	case Ttable:
	    if(p->posx>0) {
		if(p->posx+elem->spcWidth+elem->w > p->w) {
		    p->newLine();
		} else 
		    p->posx+=elem->spcWidth;
	    }
	    curTable = (Table *) elem->args;
	    // remember table position, its needed to draw the frame
	    elem->x=p->x+p->posx;
	    elem->y=p->y+p->posy;
	    curTable->curRow=0;
	    curTable->h=curTable->outerFrame+curTable->borderWidth;
	    p->posx+=curTable->outerFrame+curTable->borderWidth;
	    p->posy+=curTable->outerFrame+curTable->borderWidth;
	    // adjust table to let columns expand and take all available
	    // space
	    curTable->expand(p->w);
	    break;
	case Ttr:
	    if(curTable!=NULL) {
		curTable->curCol=0;
	    }
	    break;
	case Ttd:
	case Tth:
	    if(curTable!=NULL) {
		paraStack->push(lineDef, p);
		p->x+=curTable->innerFrame;
		p->y+=curTable->innerFrame;
		//for(int i=0; i<curTable->curCol; i++)
		//p->x+=curTable->getColWidth(i);
		//for(int i=0; i<curTable->curRow; i++)
		//p->y+=curTable->getRowHeight(i);
		p->w=curTable->getColWidth(curTable->curCol);
		//printf("<td>(%d,%d): x=%d, y=%d, w=%d\n", curTable->curRow, curTable->curCol, p->x, p->y, p->w);
	    }
	    break;
	case Utd:
	case Uth:
	    if(curTable!=NULL) {
		if(p->posx>0)
		    p->newLine();
		curTable->setRowHeight(curTable->curRow, p->posy);
		paraStack->pop(p);
		p->posx+=curTable->getColWidth(curTable->curCol)
		    +2*curTable->innerFrame+curTable->borderWidth;
		curTable->curCol++;
	    }
	    break;
	case Utr:
	    if(curTable!=NULL) {
		curTable->h+=curTable->getRowHeight(curTable->curRow)
		    +2*curTable->innerFrame+curTable->borderWidth;
		//printf("old posy=%d\n", p->posy);
		p->posy+=curTable->getRowHeight(curTable->curRow)
		    +2*curTable->innerFrame+curTable->borderWidth;
		//printf("new posy=%d\n", p->posy);
		p->posx=curTable->outerFrame+curTable->borderWidth;
		curTable->curRow++;
	    }
	    break;
	case Utable:
//	    p->posx+=curTable->w;
//		+curTable->outerFrame+curTable->borderWidth;
// 	    p->posy+=curTable->h;
//		+curTable->outerFrame+curTable->borderWidth;
	    if(curTable!=NULL) {
		// restore previous table pointer
		curTable->h+=curTable->borderWidth+curTable->outerFrame;
		
		// Store dimensions in the openeing table element, it is
		// required for clipping
		elements[curTable->elemIndex].w = curTable->w;
		elements[curTable->elemIndex].h = curTable->h;
		p->posx+=curTable->outerFrame+curTable->borderWidth;
		p->posy+=curTable->outerFrame+curTable->borderWidth;
		curTable = (Table *) elem->args;
	    }
	    break;
	}
    }
    if(p->posx>0)
	p->newLine();
    pageh=p->y+p->posy;
}

//
//
// SIZING SECTION
//
//


//
// Traverse the element list, and determine the font, font size,
// and overall width of each element. This is to determine
// the minimum screen width, among others.
//
// This sets the elem->font, elem->size, elem->spcWidth
// Sets the elem->w field
// Sets the args field of tables to the respective table
//
void HtmlData::measure() {

    Font initFont;
    initFont.font=defaultFont;
    initFont.size=defaultSize;
    Font *curFont = &initFont;
    // Set default font and compute space width
    fl_font(curFont->font, curFont->size);
    curFont->spcWidth=(uchar) fl_width(" ");

    Para *curPara = new Para(lineDef);

    Table *curTable=NULL;

    char *argVal;
    char *newTxt;

    for(int i=0; i<elemCount; i++) {
	Elem *elem=&(elements[i]);
	elem->w=0;
	switch(elem->type) {
	case Tpre:
	    // Preformatted text
//	    fontStack->push(curFont, FL_COURIER, 10);
	    elem->font=curFont->font;
	    elem->size=curFont->size;
	    elem->w=(short) fl_width(elem->args);
	    elem->h=elem->size;
	    curPara->compareWidth(elem->w);
	    break;
	case Timg:
	    // Images are handled like text, ironically
	    // See if an ALT tag is specified
	    argVal=getArg(elem, "alt");
	    if(argVal!=NULL) {
		char *replacedText=
		    replaceSpecialCharacters(argVal, true, NULL);
		free(argVal);
		argVal=replacedText;
		// yes, format it in square brackets
		newTxt=(char *) malloc(strlen(argVal)+3);
		sprintf(newTxt, "[%s]", argVal);
		// free the argument value
		free(argVal);
	    } else {
		// no, use dummy text
		newTxt=strdup("[IMAGE]");
	    }
	    // Hack: free the original argument string
	    if (elem->args==NULL)
		free(elem->args);
	    // replace it with the ALT text
	    elem->args=newTxt;
	    // From here on, Timg is treated as a synonym to Ttext!
	    fontStack->push(curFont, FL_COURIER, defaultSize);
	    elem->font=curFont->font;
	    elem->size=curFont->size;
	    elem->w=(short) fl_width(elem->args);
	    elem->h=elem->size;
	    fontStack->pop(curFont);
	    curPara->compareWidth(elem->w);
	    break;
	case Ttext:
	    // Word in a text flow
	    elem->font=curFont->font;
	    elem->size=curFont->size;
	    elem->spcWidth=curFont->spcWidth;
	    elem->w=(short) fl_width(elem->args);
	    elem->h=elem->size;
	    curPara->compareWidth(elem->w);
	    break;

	    //
	    // FONT CHANGES
	    //
	case Tb:
	case Tstrong:
	    fontStack->push(curFont, 
			   curFont->font|FL_BOLD, curFont->size);
	    break;
	case Ti:
	    fontStack->push(curFont, 
			   curFont->font|FL_ITALIC, curFont->size);
	    break;
	case Tcode:
	    fontStack->push(curFont, 
			   FL_COURIER, curFont->size);
	    break;
	case Tkbd:
	    fontStack->push(curFont, 
			   FL_COURIER_BOLD, curFont->size);
	    break;
	case Tvar:
	    fontStack->push(curFont, 
			   FL_COURIER_ITALIC, curFont->size);
	    break;
	case Tcite:
	    fontStack->push(curFont, 
			   curFont->font|FL_ITALIC, curFont->size);
	    break;
	case Th1:
	    fontStack->push(curFont,
			    defaultFont|FL_BOLD, defaultSize*200/100);
	    break;
	case Th2:
	    fontStack->push(curFont,
			    defaultFont|FL_BOLD, defaultSize*180/100);
	    break;
	case Th3:
	    fontStack->push(curFont,
			    defaultFont|FL_BOLD, defaultSize*160/100);
	    break;
	case Th4:
	    fontStack->push(curFont,
			    defaultFont|FL_BOLD, defaultSize*140/100);
	    break;
	case Th5:
	    fontStack->push(curFont,
			    defaultFont|FL_BOLD, defaultSize*120/100);
	    break;
	case Th6:
	    fontStack->push(curFont,
			    defaultFont|FL_BOLD, defaultSize*110/100);
	    break;
	case Ub:
	case Ustrong:
	case Ui:
	case Ucode:
	case Ukbd:
	case Uvar:
	case Ucite:
	case Uh1:
	case Uh2:
	case Uh3:
	case Uh4:
	case Uh5:
	case Uh6:
//	case Upre:
	    fontStack->pop(curFont);
	    break;


	    //
	    // LISTS
	    //
	case Tul:
	    paraStack->push(lineDef, curPara);
	    curPara->w=0;
	    fontStack->push(curFont, defaultFont, defaultSize);
	    elem->spcWidth=curFont->spcWidth;
	    //curPara->x+=5*curFont->spcWidth;
	    break;
	case Tol:
	    paraStack->push(lineDef, curPara);
	    curPara->w=0;
	    fontStack->push(curFont, defaultFont, defaultSize);
	    elem->spcWidth=curFont->spcWidth;
	    //curPara->x+=5*curFont->spcWidth;
	    curPara->number=1;
	    break;
	case Uul:
	case Uol:
	    // Width of list is width of biggest element
	    // plus indent
	    elem->w=curPara->w + 5*curFont->spcWidth;
	    fontStack->pop(curFont);
	    paraStack->pop(curPara);
	    // Update parent paragraph's width
	    curPara->compareWidth(elem->w);
	    break;
	case Tli:
	    elem->spcWidth=curFont->spcWidth;
	    if(curPara->number >0)
		elem->listno=curPara->nextNumber();
	    else
		elem->listno=0;
	    break;

	    //
	    // TABLES
	    //
	case Ttable:
	    // allocate a new table structure
	    tableStack->push(curTable, i);
	    // store the table descriptor with the element (HACK)
	    elem->args = (char *) curTable;
	    //curPara->x=0;
	    break;
	case Utable:
	    if(curTable!=NULL) {
		// compute minimum table width
		curTable->minW=2*curTable->outerFrame+curTable->borderWidth;
		for(int i=0; i<curTable->colCount; i++)
		    curTable->minW += curTable->getMinColWidth(i)
			+ 2*curTable->innerFrame
			+ curTable->borderWidth;
		elem->w=curTable->minW;
		//elem->h=curTable->h;
		tableStack->pop(curTable);
		//printf("table width: %d\n", elem->w);
		curPara->compareWidth(elem->w);
		// store the previous table pointer
		elem->args = (char *) curTable;
	    }
	    break;
	case Ttr:
	    if(curTable!=NULL) {
		curTable->curCol=0;
		// make sure the 'rows' array is allocated with proper size
		curTable->setRowHeight(curTable->curRow, 0);
		elem->args = (char *) curTable;
	    }
	    break;
	case Utr:
	    if(curTable!=NULL) {
		curTable->curRow++;
		elem->args = (char *) curTable;
	    }
	    break;
	case Tth:
	    if(curTable!=NULL) {
		elem->args = (char *) curTable;
		paraStack->push(lineDef, curPara);
		//curPara->x=0;
		curPara->w=0;
		fontStack->push(curFont, 
				defaultFont|FL_BOLD, defaultSize);
	    }
	    break;
	case Ttd:
	    if(curTable!=NULL) {
		elem->args = (char *) curTable;
		paraStack->push(lineDef, curPara);
		//curPara->x=0;
		curPara->w=0;
		fontStack->push(curFont, defaultFont, defaultSize);
	    }
	    break;
	case Uth:
	case Utd:
	    if(curTable!=NULL) {
		elem->args = (char *) curTable;
		// Update column widtht in table structure, and make sure
		// the 'cols' array is properly allocted, and that
		// colCount is incremented
		curTable->setMinColWidth(curTable->curCol, curPara->w);
		// pop font
		fontStack->pop(curFont);
		// pop paragraph
		paraStack->pop(curPara);
		// switch to next column
		curTable->curCol++;
		break;
	    }
	}
    }

    // pop remaining nested paras?
    // ...

    // Minimum page width
    minPageWidth=curPara->w;
}
	
	
	    



//
//
// PARSING SECTION
//
//

static TagMap tagTable[] = {
    { "!--", Tcomment},
    { "html", Thtml }, { "/html", Uhtml },
    { "head", Thead }, { "/head", Uhead },
    { "body", Tbody }, { "/body", Ubody },
    { "title", Ttitle }, { "/title", Utitle },
    { "meta", Tmeta}, { "/meta", Umeta},
    { "img", Timg}, { "/img", Uimg},
    { "br", Tbr},
    { "p", Tp},
    { "b", Tb}, { "/b", Ub},
    { "strong", Tstrong}, { "/strong", Ustrong},
    { "i", Ti}, { "/i", Ui},
    { "code", Tcode}, { "/code", Ucode},
    { "kbd", Tkbd}, { "/kbd", Ukbd},
    { "var", Tvar}, { "/var", Uvar},
    { "cite", Tcite}, { "/cite", Ucite},
    { "a", Ta}, { "/a", Ua},
    { "hr", Thr},
    { "ul", Tul}, {"/ul",Uul},
    { "ol", Tol}, {"/ol",Uol},
    { "li", Tli},
    { "table", Ttable}, {"/table", Utable},
    {"tr", Ttr}, {"/tr", Utr},
    {"th", Tth}, {"/th", Uth},
    {"td", Ttd}, {"/td", Utd},
    {"h1", Th1}, {"/h1", Uh1},
    {"h2", Th2}, {"/h2", Uh2},
    {"h3", Th3}, {"/h3", Uh3},
    {"h4", Th4}, {"/h4", Uh4},
    {"h5", Th5}, {"/h5", Uh5},
    {"h6", Th6}, {"/h6", Uh6},
    {"/pre", Upre},
    {"img", Timg}
};


char *HtmlData::getArg(Elem *elem, const char *argName) {
    char *s;
    int l, n;
    char *result=NULL;
    
    s=elem->args;
    l=strlen(argName);
    // skip over white space
    while(s!=NULL && *s!='\0') {
	while (isspace(*s)) s++;
	if(strncasecmp(s, argName, l)==0 && s[l]=='=') {
	    // Found it!
	    // skip until after the '='
	    s+=l+1;
	    n=0;
	    if(*s=='"') {
		// quoted string
		s++;
		while(s[n]!='\0' && s[n]!='"') n++;
	    } else {
		while(s[n]!='\0' && !isspace(s[n])) n++;
	    }
	    result = (char *) malloc(n+1);
	    strncpy(result, s, n);
	    result[n]='\0';
	    break;
	} else {
	    // Skip over the argument name
	    s=strchr(s, '=');
	    if(s==NULL)
		return NULL;
	    // skip over '=' sign
	    s++;
	    // Skip over the arugment value
	    if(*s=='"') {
		s++;
		// skip over quoted value
		while(*s!='\0' && *s!='"')
		    s++;
	    }
	    while(*s!='\0' && !isspace(*s)) s++;
	}
    }
    return result;
}



Elem *HtmlData::createTagElem(const char *text) {
    const char *p, *q;
    char buf[256];
    
    Elem *result = newElem();
    
    p=text;
    while(isspace(*p)) p++;
    q=p;
    while(*q!='\0' && *q!='>' && !isspace(*q)) q++;
    int n =q-p;
    if(n>=sizeof(buf)-1)
	n=sizeof(buf)-1;
    strncpy(buf, p, n);
    buf[n]='\0';

    int nTags = sizeof(tagTable)/sizeof(TagMap);
    int i;
    for(i=0; i<nTags; i++) {
	int n2=strlen(tagTable[i].name);
	if(strncasecmp(p, tagTable[i].name, n>n2? n:n2)==0) 
	    break;
    }
    if(i>=nTags) {
	result->type=Tunknown;
	//printf("Unknown tag: '%s'\n", buf);
    }
    else
	result->type=tagTable[i].type;

    while(isspace(*((char *) q))) q++;
    const char *startOfArgs=q;
    while(*q!='\0' && *q!='>')
	q++;
    if(q==startOfArgs)
	result->args=NULL;
    else {
	result->args=(char *) malloc(q-startOfArgs+1);
	strncpy(result->args, startOfArgs, q-startOfArgs);
	result->args[q-startOfArgs]='\0';
    }
    if(*q=='>')
	q++;
    result->nchars=q-text;
    return result;
}


Elem *HtmlData::createPreElem(const char *text) {
    Elem *result = newElem();

    result->type = Tpre;
    const char *p=text;
    while(*p!='\0' && *p!='\n') p++;
    result->args=(char *) malloc(p-text+1);
    strncpy(result->args, p, p-text);
    if(*p!='\0') p++;
    result->nchars=p-text;
    return result;
}

static CharMap charTable[] = {
    {"amp", '&'},
    {"lt", '<'},
    {"gt", '>'},
    {"nbsp", ' '},
    {"copy", '\251'},
    {"reg", '\256'},
    {"#132", '`'},
    {"#146", '\''},
    {"#147", '\''},
    {"#150", '-'},
    {"quot", '\"'},
    {"auml", ''},
    {"ouml", ''},
    {"uuml", ''},
    {"Auml", ''},
    {"Ouml", ''},
    {"Uuml", ''},
    {"szlig", ''}
};

char HtmlData::getSpecialChar(const char *text, int len) {
    int n = sizeof(charTable)/sizeof(CharMap);
    int i;
    for(i=0; i<n; i++) {
	if(strncmp(text, charTable[i].name, len)==0)
	    break;
    }
    if(i<n)
	return charTable[i].map;
    else
	return '?';
}

Elem *HtmlData::newElem() {

    Elem *result;
    if(elemCount>=elemAlloc) {
	elemAlloc+=50;
	if(elemAlloc==50) 
	    elements=(Elem *) malloc(sizeof(Elem) * elemAlloc);
	else 
	    elements = (Elem *) realloc(elements, sizeof(Elem)*elemAlloc);
    }
    result = elements+elemCount;
    elemCount++;
    result->lineIdx=0;
    result->x=0;
    result->y=0;
    result->args=NULL;
    result->w=0;
    result->h=0;
    result->listno=-1;
    result->linkIndex=-1;
    return result;
}


//
// This function replaces special characters like '&lt;' with the ASCII
// equivalent like '<'.
//
// The 'everything' parameter describes whether only text up to the
// next white space, or the full string is to be converted.
//
// Optional, the length of text scanned is included.
//
char *HtmlData::replaceSpecialCharacters(const char *text, bool everything,
				     int *nchars) {
    char *result;
    const char *p=text;
    char *bufp;

    //
    // Compute the # of characters that will be skipped
    //
    int skip=0;
    while(*p!='\0' && (everything || (!isspace(*p) && *p!='<'))) {
	// Look out for constructs like "&foo;"
	if(*p=='&') {
	    const char *q;
	    for(q=p; *q!='\0'&&*q!=';'&&!isspace(*q);q++);
	    if(*q!=';')q=NULL;
	    //const char *q=strchr(p, ';');
	    // sum up their length
	    if(q!=NULL)
		skip+=q-p;
	}
	p++;
    }
    result=(char *) malloc(p-text+1-skip);
    // reset pointer to start of text
    p=text;
    bufp=result;
    while(*p!='\0' && (everything || (!isspace(*p) && *p!='<'))) {
	// Look out for constructs like "&foo;"
	if(*p=='&') {
	    const char *q;
	    for(q=p; *q!='\0'&&*q!=';'&&!isspace(*q);q++);
	    if(*q!=';')q=NULL;
	    //char *q=strchr(p, ';');
	    // sum up their length
	    if(q!=NULL) {
		*bufp++=getSpecialChar(p+1, q-p-1);
		// continue after the semicolon
		p=q+1;
	    } else {
		*bufp++=*p++;
	    }
	} else 
	    *bufp++ = *p++;
    }
    *bufp='\0';
    if(nchars!=NULL)
	*nchars = p-text;
    return result;
}


Elem *HtmlData::createTextElem(const char *text) {
    Elem *result = newElem();
    
    result->type = Ttext;
    result->args=replaceSpecialCharacters(text, false, &(result->nchars));
    return result;
}

void HtmlData::findHead() {
    int length=0;
    int i, first;
    char *s;

    first=-1;
    
    if(htmlHead!=NULL) {
	free(htmlHead);
	htmlHead=NULL;
    }

    for(i=0; i<elemCount; i++) {
	Elem *elem=&(elements[i]);
	if(elem->type==Thead) {
	    first=i;
	} else if (elem->type==Uhead) {
	    if(first>0)
		break;
	} else if (elem->type==Ttext) {
	    length += strlen(elem->args)+1;
	}
    }

    if(first>=0) {
	htmlHead=(char *) malloc(length+1);
	s=htmlHead;
	for(i=first; i<elemCount && elements[i].type!=Uhead; i++) {
	    Elem *elem=&(elements[i]);
	    if(elem->type==Ttext) {
		// Insert a blank if this is not the first word
		if(s!=htmlHead)
		    *s++=' ';
		// append word to htmlHead buffer
		strcpy(s, elem->args);
		// advance s to end of buffer
		while(*s!='\0') s++;
	    }
	}
    }
}

void HtmlData::parse() {

    const char *p;
    Elem *elem;
    int preDepth=0;
    bool inComment=false;

    // Clear the element list
    elemCount=0;

    p=htmlText;
    while(p!=NULL && *p!='\0') {
	while(isspace(*p))
	    p++;
	// Handle comments
	if(strncmp(p, "<!--", 4)==0) {
	    p+=4;
	    while(*p!='\0' && strncmp(p, "-->", 3)!=0) p++;
	    if(*p!='\0')
		p+=3;
	    continue;
	}

	if (strncasecmp(p, "</pre>", 6)==0 && preDepth>0) {
	    preDepth--;
	}
	if (preDepth>0) {
	    elem=createPreElem(p);
	    p+=elem->nchars;
	} else 	if (strncasecmp(p, "<pre>", 5)==0) {
	    p+=5;
	    preDepth++;
	} else if(*p=='<') {
	    p++;
	    elem=createTagElem(p);
	    p+=elem->nchars;
	} else {
	    elem=createTextElem(p);
	    p+=elem->nchars;
	}
    }
    // Can free memory for unparsed text now
    free((void *)htmlText);
    htmlText=NULL;

    // find HEAD string
    findHead();

}


void HtmlData::dump() {
    int i;
    //printf("Have %d elements\n", elemCount);
    int extra=0;
    for(i=0; i<elemCount; i++) {
	Elem *elem=&elements[i];
	int type=elem->type;
	int nTags = sizeof(tagTable)/sizeof(TagMap);
	int j;
	for(j=0; j<nTags; j++) {
	    if(type==tagTable[j].type)
		break;
	}
	if(j>=nTags)
	    printf("type: %d\n", elem->type);
	else
	    printf("type: %d (%s)\n", elem->type, tagTable[j].name);
	printf("args: %s\n", elem->args);
	printf("nchr: %d\n\n", elem->nchars);
    }
    printf("Size of elements array: %d\n\n", elemAlloc*sizeof(Elem));
}


void HtmlData::clear() {
    if(htmlText!=NULL) {
	free((void*) htmlText);
	htmlText=NULL;
    }
    if(htmlHead!=NULL) {
	free(htmlHead);
	htmlHead=NULL;
    }
    fontStack->clear();
    paraStack->clear();
    tableStack->clear();
    lineDef->clear();
    linkArray->clear();
    // Note: we're leaking most of the things pointed to by the
    // elements here!!!
    elemCount=0;
}

HtmlData::HtmlData() {
    elemCount=0;
    elemAlloc=0;

    defaultFont=FL_HELVETICA;
#ifdef HOST
    defaultSize=12;
#else
    defaultSize=8;
#endif

//    fontCount=0;
//    fontAlloc=0;
//    paraCount=0;
//    paraAlloc=0;
//    linkCount=0;
//    linkAlloc=0;
    fontStack = new(FontStack);
    paraStack = new(ParaStack);
    tableStack = new (TableStack);
    lineDef = new(LineDef);
    linkArray = new(LinkArray);
    htmlText=NULL;
    htmlHead=NULL;
}


HtmlData::~HtmlData() {
    clear();
    delete(fontStack);
    delete(paraStack);
    delete(tableStack);
    delete(lineDef);
}

