//
// Copyright (C) 2002 Henning Spruth

// Written by Henning Spruth <henning.spruth@epost.de>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


typedef enum {
    Ttext,	//0
    Tpre,
    Tcomment,
    Thtml, Uhtml,
    Thead, Uhead,
    Tbody, Ubody,
    Ttitle, Utitle,
    Tmeta, Umeta,
    Timg, Uimg,
    Tbr,
    Tp,
    Tb, Ub,
    Ti, Ui,
    Tcode, Ucode,
    Tkbd, Ukbd,
    Tvar, Uvar,
    Tcite, Ucite,
    Tstrong, Ustrong,
    Ta, Ua,
    Thr,
    Tul, Uul,
    Tol, Uol,
    Tli,
    Ttable, Utable,
    Ttr, Utr,
    Tth, Uth,
    Ttd, Utd,
    Th1, Uh1,
    Th2, Uh2,
    Th3, Uh3,
    Th4, Uh4,
    Th5, Uh5,
    Th6, Uh6,
    Upre,
    Tunknown 
} ElemType;


//
// Structure for the mapping of HTML tag strings to the enums
// defined above, e.g. "<table>" -> Ttable
//
typedef struct {
    const char *name;
    ElemType type;
} TagMap;


//
// Structure for the mapping of HTML character esacapes to ASCII,
// e.g. "&amp;" -> "&"
//
typedef struct {
    const char *name;
    const char map;
} CharMap;




typedef struct {
    ElemType type;
    int nchars;	         // # of characters this element took in the source
    char *args;	
    short lineIdx;	// pointer to structure representing output line
    uchar font;		// Text font
    uchar  size;		// Text size
    uchar spcWidth;
    char border;		// Draw border?
    short x;		// Indentation/starting X coordinate
    short y;		// Starting Y coordinate
    short w;		// Width
    short h;		// Height
    short listno;	// Number for a numbered list
    short linkIndex;	// Index into link table

} Elem;


//typedef struct {
//    char font;
//    char size;
//} DisplayFont;

//typedef struct {
//    short indent;
//    short number;
//} DisplayPara;

class HtmlData {

private:
    // The raw HTML read in
    const char *htmlText;

    // The text from the HTLL HEAD tag
    char *htmlHead;

    // functions helping in creating the parse tree
    Elem *createTagElem(const char *text);
    Elem *createPreElem(const char *text);
    char getSpecialChar(const char *text, int len);
    Elem *createTextElem(const char *text);
    Elem *newElem();

    // allocates a new Link structure, and returns its index
    short newLinkIndex(char *link);

    // management of fonts and paragraphs for the render phase
    void pushFont(uchar &font, uchar &size, 
		  uchar newfont, uchar newsize, uchar &spcwidth);
    void popFont(uchar &font, uchar &size, uchar &spcwidth);
    void pushPara(short &indent, short newindent, short number);
    void popPara(short &indent);

    short pagew;	// current width of HTML page
    short pageh;	// current height of HTML page
    short minPageWidth; // minimum widht of page

    uchar defaultFont;
    uchar defaultSize;


    // managing elements
    int elemCount;
    int elemAlloc;
    Elem *elements;

//    // managing font stack
//    int fontCount;
//    int fontAlloc;
//    DisplayFont *fonts;

//    //managing links
//    int linkCount;
//    int linkAlloc;
//    Link *links;

//    // managing paragraphs
//    int paraCount;
//    int paraAlloc;
//    DisplayPara *paras;

    // managing tables
    FontStack *fontStack;
    ParaStack *paraStack;
    TableStack *tableStack;
    LineDef *lineDef;
    LinkArray *linkArray;

    void findHead();

public:
    void loadFile(const char *fileName);
    void parse();
    void measure();
    void format(int screenWidth);
    void dump();
    void clear();
    int getElemCount() { return elemCount; };
    Elem *getElem(int i) { return &elements[i]; };
    short getElemMaxSize(Elem *);
    short getPageW() {return pagew;};
    short getPageH() {return pageh;};
    char *getArg(Elem *elem, const char *argName);
    char *replaceSpecialCharacters(const char *text, bool everything,
				   int *nchars);
    
    Link *getLink(short index);
    const char *getHead() {return htmlHead; }

    HtmlData();
    ~HtmlData();


};
