//
// Copyright (C) 2002 Henning Spruth

// Written by Henning Spruth <henning.spruth@epost.de>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

//
// A LineDef manages an array of 'Line' structures.
//
// Each 'Line' structure stores data for one line of an output
// paragraph, currently this is only to line height (maximum
// font size of all characters)
//


#include <stdlib.h>
#include <assert.h>

#ifdef DMALLOC
#include <dmalloc.h>
#endif

#include "LineDef.h"



LineDef::LineDef() {
    alloc=increment;
    count=0;
    stack=(Line *) calloc(alloc, sizeof(Line));
//    stack=(Line *) malloc(alloc* sizeof(Line));
}

LineDef::~LineDef() {
    free((void *) stack);
    stack=NULL;
}


int LineDef::newLine() {
    if(count>=alloc) {
	alloc+=increment;
	Line *old=stack;
	stack = (Line *) realloc(stack, sizeof(Line)*alloc);
	//assert(old==stack);
    }
    count++;
    stack[count-1].maxSize=0;
    return count-1;
}

void LineDef::clear() {
    count=0;
}

short LineDef::maxSize(int i) {
    assert(i<count);
    return stack[i].maxSize;
}

void LineDef::testSize(int i, short size) {
    assert(i<count);
    if(stack[i].maxSize<size)
	stack[i].maxSize=size;
}

