//
// Copyright (C) 2002 Henning Spruth

// Written by Henning Spruth <henning.spruth@epost.de>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#ifdef DMALLOC
#include <dmalloc.h>
#endif


#include "Url.h"

int Url::isImplicitHttp(char *string) {
    int have_dot=0;
    int have_char=0;
    char *s=string;
    // Must start with an alphanumeric character
    if(!isalnum(*s))
	return 0;
    // look at every character
    while(*s!='\0') {
	if(*s=='.')
	    have_dot=1;
	else if(!isalnum(*s))
	    break;
	have_char=1;
	s++;
    }
    // minimum length is 5
    if(s-string <5)
	return 0;
    // must containg both '.' and an alphanumeric
    if(!have_dot || !have_char)
	return 0;
    // must not end in a '.'
    if(*(s-1)=='.')
	return 0;
    // Seems to be a HTTP address
    return 1;
}

char *Url::newDirectoryPartOf(char *path) {
    char *s = path+strlen(path);
    // search for the trailing slash
    while(s>path && *s!='/')
	s--;
    // Exclude trailing slash
    if(s>path && *s=='/')
	s--;
    // Allocate result and copy directory information
    char *result = (char *) malloc(s-path+2);
    strncpy(result, path, s-path+1);
    result[s-path+1]='\0';
    return result;

}

void Url::setTypeAndHost() {
    char *start=url;
    if(strncmp(url, "http://", 7)==0) {
	type=URL_HTTP;
	implicit=false;
	start=url+7;
    } else if (strncmp(url, "file:",5)==0) {
	type=URL_FILE;
	implicit=false;
	start=url+5;
    } else if (strncmp(url, "zip:",4)==0) {
	type=URL_ZIP;
	implicit=false;
	start=url+4;
    } else if(isImplicitHttp(url)) {
	type=URL_HTTP;
	implicit=true;
    } else {
	type=URL_FILE;
	implicit=true;
    }

    if(type!=URL_FILE) {
	// Note: for URL_ZIP the leaf name of the zip file
	// is used as the host name ...
	char *hostEnd=start;
	while(*hostEnd!='\0' && *hostEnd!='/')
	    hostEnd++;

	// get host part
	int n=hostEnd-start;
	host=(char *) malloc(n+1);
	strncpy(host, start, n);
	host[n]='\0';
	
	// get path part
	path=strdup(hostEnd);
    } else {
	host=NULL;
	path=strdup(start);
    }
    dir=newDirectoryPartOf(path);
}

Url::Url(const char *_url) {
    strcpy(url, _url);
    host=path=dir=NULL;
    setTypeAndHost();
}

Url::Url() {
    url=host=path=dir=NULL;
}

void Url::set(const char *_url) {
    clear();
    url=strdup(_url);
    setTypeAndHost();

}

void Url::clear() {
    if(url!=NULL) free(url);
    if(host!=NULL) free(host);
    if(path!=NULL) free(path);
    if(dir!=NULL) free(dir);
    url=host=path=dir=NULL;
}


Url::~Url() {
    clear();
}

