//
// Copyright (C) 2002 Henning Spruth

// Written by Henning Spruth <henning.spruth@epost.de>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <stdlib.h>
#include <stdio.h>
#ifdef DMALLOC
#include <dmalloc.h>
#endif

#include <FL/Fl.H>
#include <FL/fl_draw.H>

#include "LinkArray.h"
#include "FontStack.h"
#include "LineDef.h"
#include "Para.h"
#include "ParaStack.h"
#include "Table.h"
#include "TableStack.h"
#include "HtmlData.h"
#include "VisLinkArray.h"

VisLinkArray::VisLinkArray() {
    stack=(LinkBox *) calloc(increment, sizeof(LinkBox));
    alloc=increment;
    count=0;
}

VisLinkArray::~VisLinkArray() {
    free((void *) stack);
}

void VisLinkArray::clear() {
    // reset count to 0
    count=0;
    // NOTE: the Link members are **not** freed as its assumed that this
    // is taken care of by LinkArray::clear()
}

void VisLinkArray::add(short x, short y, short w, short h, Link *link) {
    //printf("add: link target %s\n", link->target);
    if(count>=alloc) {
	alloc+=increment;
	stack = (LinkBox *) realloc(stack, sizeof(LinkBox)*alloc);
    }
    LinkBox *lb = &stack[count++];
    lb->x=x;
    lb->y=y;
    lb->w=w;
    lb->h=h;
    lb->link=link;
}

Link *VisLinkArray::underCursor(short cx, short cy) {
    LinkBox *lb;
    lb=stack;
    for(int i=0; i<count; i++, lb++) {
	if(cx>=lb->x && cx<lb->x+lb->w && cy>=lb->y && cy<lb->y+lb->h) {
	    return lb->link;
	}
    }
    return NULL;
}
